"""
TODO:
[ ] Implement segmented heap to handle heterogeneous hardware resources.

TODO (BACKLOG):
[ ] Handle many connections. 
    Probably want to put the server into some kind of not ready state.
    Can possibly implement this behavior using gRPC service intercept handler.
[ ] NEED to write tests to identify bugs in my shitty multithreading code.
[ ] Graceful handling of endlessly spinning jobs. Possibly implement a timeout?
[ ] Continuous job profiling might be useful as a way to gracefully handle interruptions / failures.
    That way, there doesn't need to be special logic in the client to send a message when abnormal
    termination happens.
    Implement this using a gRPC stream.
"""
import sys
import os
# import time
from concurrent import futures
# import threading
# from threading import Thread
import signal
import logging

import grpc

from grn.utils.utils import find_free_port
from grn.utils.gpu_monitors import GPUMonitor
import grn.grpc_resources.master_pb2 as protos
import grn.grpc_resources.master_pb2_grpc as services
from grn.core.constants import GRN_SERVER_TMP_INFO_PATH, ServiceErrorCode, ResourcePolicy
from grn.core.globals import GPUStates, JobStates

from typing import Optional, Callable


log = logging.getLogger(__file__)


__all__ = ['serve']


# class BackgroundEventTrigger(Thread):
#     """
#     """
#     def __init__(self, event: threading.Event, delay=5):
#         super().__init__()
#         self.event = event
#         self.stopped = False
#         self.delay = delay
#         self.start()

#     def run(self):
#         while not self.stopped:
#             time.sleep(self.delay)
#             self.event.set()

#     def __enter__(self):
#         return self

#     def __exit__(self, type, value, traceback):
#         self.stopped = True


def get_next_available_gpu(jobstr: str, resource_policy: ResourcePolicy) -> protos.GPU:
    GPUStates.STATES_INIT.wait()
    is_new_job_type = (JobStates.JOB_TYPES.get(jobstr) is None)

    if len(GPUStates.GPU_QUEUE) == 0:
        errorcode = ServiceErrorCode.WAITING_FOR_JOB_PROFILE

    else:
        with GPUStates.LOCK:
            # FIXME: We assume GPU homogeneity -- all GPUs have identical architecture.
            # TODO: Implement a segmented heap to group homogenous hardware together.
            #       Then we can have a resource heap for each bin.

            def _find_first_non_profiling_gpu(pop_func: Callable) -> int:
                num_gpus = len(GPUStates.GPU_QUEUE)
                i = 0
                gid = None
                while (i < num_gpus) and len(GPUStates.GPU_QUEUE):
                    gid = pop_func()[0]
                    if gid not in GPUStates.PROFILING_GROUP:
                        break
                    i += 1
                else:
                    return None
                return gid


            if is_new_job_type:
                gid = _find_first_non_profiling_gpu(GPUStates.GPU_QUEUE.poplast)
            else:
                if resource_policy == ResourcePolicy.SPREAD:
                    gid = _find_first_non_profiling_gpu(GPUStates.GPU_QUEUE.poplast)
                elif resource_policy == ResourcePolicy.PACK:
                    gid = _find_first_non_profiling_gpu(GPUStates.GPU_QUEUE.popfirst)
                else:
                    raise NotImplementedError(resource_policy)


            def _handle_pack_policy(gid: int, mem_request: int) -> Optional[int]:
                i = 0
                group = [ (gid, GPUStates.STATES[gid]['memoryUsed']) ]
                while (i < len(GPUStates.GPU_QUEUE)):
                    if (GPUStates.STATES[group[-1][0]]['memoryFree'] >= mem_request):
                        break
                    group.append( GPUStates.GPU_QUEUE.popfirst() )

                else:
                    GPUStates.GPU_QUEUE.extend(group)
                    return None

                # Reinsert all but last element; last element will be given new resource priority.
                GPUStates.GPU_QUEUE.extend(group[:-1])  
                return group[-1][0]
            

            def _handle_spread_policy(gid: int, mem_request: int) -> Optional[int]:
                if GPUStates.STATES[gid]['memoryFree'] < mem_request:
                    return None
                return gid


            # If we reached the end of the heap, there must be no devices that aren't locked.
            if gid is None:
                errorcode = ServiceErrorCode.WAITING_FOR_JOB_PROFILE
            
            else:
                mem_free = GPUStates.STATES[gid]['memoryFree']
                mem_total = GPUStates.STATES[gid]['memoryTotal']
                # NOTE: Single python dict op should be inherently thread-safe.
                requested_memory = JobStates.JOB_TYPES.get(jobstr)

                log.debug(f'[get_next_available_gpu] (total) {mem_total}, (free) {mem_free}, (request) {requested_memory}')

                errorcode = ServiceErrorCode.OK
                if requested_memory is not None:
                    if resource_policy == ResourcePolicy.SPREAD:
                        gid = _handle_spread_policy(gid, requested_memory)
                    elif resource_policy == ResourcePolicy.PACK:
                        gid = _handle_pack_policy(gid, requested_memory)
                    else:
                        raise NotImplementedError(resource_policy)
                    
                    if gid is None:
                        errorcode = ServiceErrorCode.EXCEEDS_CURRENT_MEMORY
                    else:
                        # Reinsert gpu with new resource priority.
                        # We need to update GPU state right now so that other threads
                        # will immediately become aware of new resource constraints.
                        # Also need to update the heap so that other threads don't select
                        # the GPU we just filled.
                        GPUStates.STATES[gid]['memoryUsed'] += requested_memory
                        GPUStates.STATES[gid]['memoryFree'] -= requested_memory
                        GPUStates.GPU_QUEUE.insert(gid, GPUStates.STATES[gid]['memoryUsed'])

                # Check for other instances of this unprofiled job type.
                # We shouldn't launch this job until we get a profile.
                # NOTE: Single dict op should be inherently thread-safe.
                elif JobStates.ACTIVE_JOBS.get(jobstr):
                    errorcode = ServiceErrorCode.WAITING_FOR_JOB_PROFILE
                
                # Artificially mark this device as fully allocated to avoid running
                # other new profile jobs on it.
                elif is_new_job_type:
                    GPUStates.STATES[gid]['memoryFree'] = 0

                    # Add a device lock that will persist until a profile is sent back.
                    # This is to prevent this device from being used for any other jobs
                    # since we don't know how many resources are required yet.
                    # NOTE: This device lock can only removed in the CompleteJob service.
                    GPUStates.PROFILING_GROUP[gid] = jobstr

    # If anything went wrong, set the GPU ID to something unreachable.
    if errorcode != ServiceErrorCode.OK:
        gid = -1

    return protos.GPU(
        gpu_id=gid, 
        errorcode=errorcode.value)


# TODO: These need to be methods in grn.core.globals
def update_job_state(jobstr: str, state) -> None:
    # Accumulate the max resource consumption.
    with JobStates.LOCK:
        JobStates.JOB_TYPES[jobstr] = max(
            JobStates.JOB_TYPES.get(jobstr, state),
            state)


def push_active_job(jobstr: str) -> None:
    with JobStates.LOCK:
        JobStates.ACTIVE_JOBS[jobstr] = JobStates.ACTIVE_JOBS.get(jobstr, 0) + 1


def pop_active_job(jobstr: str) -> None:
    with JobStates.LOCK:
        JobStates.ACTIVE_JOBS[jobstr] = JobStates.ACTIVE_JOBS.get(jobstr, 1) - 1


class GPUMasterServicer(services.GPUMasterServicer):
    """
    """
    def __init__(self, max_num_jobs: Optional[int] = None):
        # NOTE: Don't really need to clear these, but doing so provides an explicit
        #       in-code reminder that GPUMasterServicer uses these globals.
        JobStates.JOB_TYPES.clear()
        JobStates.ACTIVE_JOBS.clear()
        if (max_num_jobs is not None) and (max_num_jobs <= 0):
            raise ValueError('max_num_jobs must be a positive integer')
        JobStates.MAX_NUM_JOBS = max_num_jobs

    def RequestGPU(self, request: protos.JobType, context) -> protos.GPU:
        """Service that client invokes to obtain GPU resources.

        If this is a new job type, the GPU ID provisioned is guaranteed 
        to not have another profiling job running on it.

        If this is a profiled job type, then a GPU ID will be doled out
        according to the cached resource profile. If the resource request
        exceeds available resources, then wait until the request can be
        fulfilled.
        """
        log.debug(f'[RequestGPU] request\n{request}')
        jobstr = request.jobstr
        resource_policy = ResourcePolicy(request.resource_policy)
        gpu = get_next_available_gpu(jobstr, resource_policy)
        
        errorcode = ServiceErrorCode(gpu.errorcode)
        if errorcode == ServiceErrorCode.OK:
            push_active_job(jobstr)
            log.debug(f'[RequestGPU] serving GPU ID {gpu.gpu_id}')

        return gpu

    def CompleteJob(self, request: protos.JobProfile, context) -> protos.Empty:
        """Service that a client uses to send a completed job profile.
        If this is a new job type, then profile results will be cached
        and used to allocate resources to future jobs of this type.
        """
        log.debug(f'[CompleteJob] request\n{request}')
        jobstr = request.jobtype.jobstr
        # NOTE: Single python dict ops should be inherently thread-safe, so this
        #       should be okay.
        is_new_job_type = (JobStates.JOB_TYPES.get(jobstr) is None)

        # # FIXME: There is a problem where the SingleGPUMonitor will report the full
        # #        GPU usage across all jobs, which can lead to incorrect estimates of
        # #        resource consumption for job types.
        # #        As such, do not update job state continuously.
        # #        This means we have to rely on accurate profiling of the first job.
        # if request.succeeded:
        #     update_job_state(jobstr, state=request.max_gpu_memory_used)
        pop_active_job(jobstr)

        # When a job receives a profile, we should signal threads that are waiting on 
        # job profiles to attempt to resolve their resource requests.
        if is_new_job_type and request.succeeded:
            # Remove the device lock since profiling has finished.
            assert request.gpu.gpu_id in GPUStates.PROFILING_GROUP
            assert GPUStates.PROFILING_GROUP[request.gpu.gpu_id] == jobstr
            del GPUStates.PROFILING_GROUP[request.gpu.gpu_id]

            update_job_state(jobstr, state=request.max_gpu_memory_used)

        return protos.Empty()

    def JobTypeExists(self, request: protos.JobType, context) -> protos.BoolResponse:
        """Service that simply checks if a job type has already been profiled.
        """
        return protos.BoolResponse(
            response=(JobStates.JOB_TYPES.get(request.jobstr) is not None))


def __server_shutdown_sig_handler(*args) -> None:
    log.info('[GPUMaster] Cleaning up...')
    if os.path.exists(GRN_SERVER_TMP_INFO_PATH):
        os.remove(GRN_SERVER_TMP_INFO_PATH)
    sys.exit()


def serve(debug=False, max_workers=10):
    # TODO: Implement multiple servers at once without needing separate system environments (e.g. virtual environment, docker container).
    if os.path.isfile(GRN_SERVER_TMP_INFO_PATH):
        raise SystemError(
            f'GPU Master is already running! Shut down current server process before launching a new one.')

    signal.signal(signal.SIGINT, __server_shutdown_sig_handler)
    signal.signal(signal.SIGTERM, __server_shutdown_sig_handler)

    # Obtain a random free port from the OS and cache it to a secret file.
    # This file will be removed once the server shuts down.
    port = find_free_port()
    with open(GRN_SERVER_TMP_INFO_PATH, 'w') as grn_file:
        grn_file.write(f'{port}')

    with GPUMonitor(delay=0.1):
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
        services.add_GPUMasterServicer_to_server(GPUMasterServicer(), server)
        server.add_insecure_port(f'[::]:{port}')
        server.start()
        server.wait_for_termination()

    # Just in case.
    __server_shutdown_sig_handler()


import os

import grpc

from grn.core.constants import GRN_SERVER_TMP_INFO_PATH


__all__ = ['find_free_port', 'is_server_available']


def find_free_port():
    import socket
    from contextlib import closing
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind(('', 0))
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return sock.getsockname()[1]


def is_server_available(address: str = 'localhost:50051') -> bool:
    channel = grpc.insecure_channel(address, options=(('grpc.enable_http_proxy', 0),))
    try:
        grpc.channel_ready_future(channel).result(timeout=1)
    except grpc.FutureTimeoutError:
        return False
    return True


def find_gpu_master_address() -> str:
    # Find the GPU Master port
    if os.path.isfile(GRN_SERVER_TMP_INFO_PATH):
        with open(GRN_SERVER_TMP_INFO_PATH, 'r') as grn_file:
            gpu_master_addr = f'localhost:{grn_file.read()}'
    # File containing current server port must exist.
    else:
        raise FileNotFoundError(f'{GRN_SERVER_TMP_INFO_PATH}, serve never called.')
    return gpu_master_addr



from setuptools import setup, find_packages

install_requires = [
    'nvidia-ml-py3',
    'grpcio',
    'depq',
    'rich',
]

dev_requires = [
    'grpcio-tools',
]

setup(
    name='gpu-resource-negotiator',
    packages=find_packages(),
    version='0.2.3',
    license='unlicense',
    description='GPU Resource Negotiator (GRN). Map N jobs to M devices on the same machine.',
    author='Jonathan Helland',
    url='https://github.com/j-helland/grn',
    download_url='https://github.com/j-helland/grn/archive/refs/tags/pre-v-0.2.3.tar.gz',
    keywords=['GPU', 'load balancing'],
    install_requires=install_requires,
    extras_require={ 'dev': dev_requires },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: The Unlicense (Unlicense)',
        'Programming Language :: Python :: 3',
    ],
    scripts=['bin/grn-start', 'bin/grn-run']
)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: master.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='master.proto',
  package='grn',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0cmaster.proto\x12\x03grn\"\x07\n\x05\x45mpty\" \n\x0c\x42oolResponse\x12\x10\n\x08response\x18\x01 \x01(\x08\"2\n\x07JobType\x12\x0e\n\x06jobstr\x18\x01 \x01(\t\x12\x17\n\x0fresource_policy\x18\x02 \x01(\x05\"(\n\x03GPU\x12\x0e\n\x06gpu_id\x18\x01 \x01(\x05\x12\x11\n\terrorcode\x18\x02 \x01(\x05\"r\n\nJobProfile\x12\x1d\n\x07jobtype\x18\x01 \x01(\x0b\x32\x0c.grn.JobType\x12\x15\n\x03gpu\x18\x02 \x01(\x0b\x32\x08.grn.GPU\x12\x11\n\tsucceeded\x18\x03 \x01(\x08\x12\x1b\n\x13max_gpu_memory_used\x18\x04 \x01(\x03\x32\x95\x01\n\tGPUMaster\x12&\n\nRequestGPU\x12\x0c.grn.JobType\x1a\x08.grn.GPU\"\x00\x12,\n\x0b\x43ompleteJob\x12\x0f.grn.JobProfile\x1a\n.grn.Empty\"\x00\x12\x32\n\rJobTypeExists\x12\x0c.grn.JobType\x1a\x11.grn.BoolResponse\"\x00\x62\x06proto3'
)




_EMPTY = _descriptor.Descriptor(
  name='Empty',
  full_name='grn.Empty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21,
  serialized_end=28,
)


_BOOLRESPONSE = _descriptor.Descriptor(
  name='BoolResponse',
  full_name='grn.BoolResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='response', full_name='grn.BoolResponse.response', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30,
  serialized_end=62,
)


_JOBTYPE = _descriptor.Descriptor(
  name='JobType',
  full_name='grn.JobType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='jobstr', full_name='grn.JobType.jobstr', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_policy', full_name='grn.JobType.resource_policy', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=64,
  serialized_end=114,
)


_GPU = _descriptor.Descriptor(
  name='GPU',
  full_name='grn.GPU',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='gpu_id', full_name='grn.GPU.gpu_id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='errorcode', full_name='grn.GPU.errorcode', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=116,
  serialized_end=156,
)


_JOBPROFILE = _descriptor.Descriptor(
  name='JobProfile',
  full_name='grn.JobProfile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='jobtype', full_name='grn.JobProfile.jobtype', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gpu', full_name='grn.JobProfile.gpu', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='succeeded', full_name='grn.JobProfile.succeeded', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_gpu_memory_used', full_name='grn.JobProfile.max_gpu_memory_used', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=158,
  serialized_end=272,
)

_JOBPROFILE.fields_by_name['jobtype'].message_type = _JOBTYPE
_JOBPROFILE.fields_by_name['gpu'].message_type = _GPU
DESCRIPTOR.message_types_by_name['Empty'] = _EMPTY
DESCRIPTOR.message_types_by_name['BoolResponse'] = _BOOLRESPONSE
DESCRIPTOR.message_types_by_name['JobType'] = _JOBTYPE
DESCRIPTOR.message_types_by_name['GPU'] = _GPU
DESCRIPTOR.message_types_by_name['JobProfile'] = _JOBPROFILE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Empty = _reflection.GeneratedProtocolMessageType('Empty', (_message.Message,), {
  'DESCRIPTOR' : _EMPTY,
  '__module__' : 'master_pb2'
  # @@protoc_insertion_point(class_scope:grn.Empty)
  })
_sym_db.RegisterMessage(Empty)

BoolResponse = _reflection.GeneratedProtocolMessageType('BoolResponse', (_message.Message,), {
  'DESCRIPTOR' : _BOOLRESPONSE,
  '__module__' : 'master_pb2'
  # @@protoc_insertion_point(class_scope:grn.BoolResponse)
  })
_sym_db.RegisterMessage(BoolResponse)

JobType = _reflection.GeneratedProtocolMessageType('JobType', (_message.Message,), {
  'DESCRIPTOR' : _JOBTYPE,
  '__module__' : 'master_pb2'
  # @@protoc_insertion_point(class_scope:grn.JobType)
  })
_sym_db.RegisterMessage(JobType)

GPU = _reflection.GeneratedProtocolMessageType('GPU', (_message.Message,), {
  'DESCRIPTOR' : _GPU,
  '__module__' : 'master_pb2'
  # @@protoc_insertion_point(class_scope:grn.GPU)
  })
_sym_db.RegisterMessage(GPU)

JobProfile = _reflection.GeneratedProtocolMessageType('JobProfile', (_message.Message,), {
  'DESCRIPTOR' : _JOBPROFILE,
  '__module__' : 'master_pb2'
  # @@protoc_insertion_point(class_scope:grn.JobProfile)
  })
_sym_db.RegisterMessage(JobProfile)



_GPUMASTER = _descriptor.ServiceDescriptor(
  name='GPUMaster',
  full_name='grn.GPUMaster',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=275,
  serialized_end=424,
  methods=[
  _descriptor.MethodDescriptor(
    name='RequestGPU',
    full_name='grn.GPUMaster.RequestGPU',
    index=0,
    containing_service=None,
    input_type=_JOBTYPE,
    output_type=_GPU,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CompleteJob',
    full_name='grn.GPUMaster.CompleteJob',
    index=1,
    containing_service=None,
    input_type=_JOBPROFILE,
    output_type=_EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='JobTypeExists',
    full_name='grn.GPUMaster.JobTypeExists',
    index=2,
    containing_service=None,
    input_type=_JOBTYPE,
    output_type=_BOOLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_GPUMASTER)

DESCRIPTOR.services_by_name['GPUMaster'] = _GPUMASTER

# @@protoc_insertion_point(module_scope)
